module GeometryTests

using Test
using Gridap
using Gridap.FieldValues
using Gridap.CellValues
using Gridap.CellValues.ConstantCellValues
using Gridap.CellMaps
using Gridap.Geometry
using Gridap.Geometry.Cartesian
using Gridap.Geometry.Unstructured
using Gridap.Geometry.Wrappers
using Gridap.Polytopes
using Gridap.Vtkio

@testset "CartesianGrid" begin

  grid = CartesianGrid(domain=(0.0,1.0,-1.0,2.0),partition=(3,4))

  @test celldim(grid) == 2
  @test pointdim(grid) == 2
  @test ncells(grid) == 3*4
  @test npoints(grid) == (3+1)*(4+1)

  x = points(grid)

  @test isa(x,CellValue{Point{2}})

  @test length(x) == 20
  @test x[13] == [0.0,1.25]
  @test x[4] == [1.0,-1.0]
  @test x[1] == [0.0,-1.0]
  @test x[end] == [1.0,2.0]

  t = cells(grid)

  @test isa(t,CellArray{Int,1})

  @test length(t) == 12

  @test t[1] == [1,2,5,6]
  @test t[11] == [14,15,18,19]

  c = celltypes(grid)
  @test isa(c,ConstantCellValue{NTuple{2,Int}})
  @test celldata(c) == (HEX_AXIS,HEX_AXIS)
  @test length(c) == 12

  o = cellorders(grid)
  @test isa(o,ConstantCellValue{Int})
  @test celldata(o) == 1
  @test length(o) == 12

  graph = gridgraph(grid)

  @test isa(graph,GridGraph)

  @test isa(celltovefs(graph), IndexCellArray{Int,1})

  @test isa(veftocells(graph), IndexCellArray{Int,1})

  trian = triangulation(grid)

  xe = cellcoordinates(trian)
  @test isa(xe,CellPoints{2})

  cb = cellbasis(trian)
  @test isa(cb,CellBasis{2,ScalarValue})

end

@testset "FlexibleUnstructuredGrid" begin

  cgrid = CartesianGrid(domain=(0.0,1.0,-1.0,2.0),partition=(3,4))

  grid = FlexibleUnstructuredGrid(cgrid)

  c = celltypes(grid)

  @test isa(c,ConstantCellValue{NTuple{2,Int}})
  @test celldata(c) == (HEX_AXIS,HEX_AXIS)
  @test length(c) == 12

  o = cellorders(grid)
  @test isa(o,ConstantCellValue{Int})
  @test celldata(o) == 1
  @test length(o) == 12

  d = mktempdir()
  f = joinpath(d,"grid")

  writevtk(grid,f)

  rm(d,recursive=true)

end

@testset "UnstructuredGrid" begin

  cgrid = CartesianGrid(domain=(0.0,1.0,-1.0,2.0),partition=(3,4))

  grid = UnstructuredGrid(cgrid)

  c = celltypes(grid)
  @test isa(c,ConstantCellValue{NTuple{2,Int}})
  @test celldata(c) == (HEX_AXIS,HEX_AXIS)
  @test length(c) == 12

  o = cellorders(grid)
  @test isa(o,ConstantCellValue{Int})
  @test celldata(o) == 1
  @test length(o) == 12

  d = mktempdir()
  f = joinpath(d,"grid")

  writevtk(grid,f)

  rm(d,recursive=true)

end

@testset "FaceLabels" begin

  vertex_to_geolabel = [1,1,2,2,2,1,1,3,3]
  edge_to_geolabel = [4,4,5,5,5,5,6,6,4]
  physlabel_1 = [1,3,4]
  physlabel_2 = [5,3,6,2]
  tag_to_name = ["label1","label2"]

  labels = FaceLabels(
    [vertex_to_geolabel, edge_to_geolabel],
    [physlabel_1, physlabel_2],
    tag_to_name)

  @test isa(labels,FaceLabels)
  @test labels_on_dim(labels,0) == vertex_to_geolabel
  @test labels_on_dim(labels,1) == edge_to_geolabel
  @test labels_on_tag(labels,1) == physlabel_1
  @test labels_on_tag(labels,2) == physlabel_2
  @test tag_from_name(labels,"label1")==1

end

@testset "Wrappers" begin

  polytope = Polytope((1,1,1))

  refcell = RefCell(polytope)

  cgrid = CartesianGrid(
    domain=(0.0,1.0,-1.0,2.0,0.0,1.0),
    partition=(3,4,2))

  grid = UnstructuredGrid(cgrid)

  ugrid = UGrid(grid)

  graph = FullGridGraph(grid)
  @test isa(connections(graph,3,0), CellArray)
  @test isa(connections(graph,3,1), CellArray)
  @test isa(connections(graph,3,2), CellArray)
  @test isa(connections(graph,2,0), CellArray)
  @test isa(connections(graph,2,1), CellArray)
  @test isa(connections(graph,2,3), CellArray)
  @test isa(connections(graph,1,0), CellArray)
  @test isa(connections(graph,1,2), CellArray)
  @test isa(connections(graph,1,3), CellArray)
  @test isa(connections(graph,0,0), CellArray)
  @test isa(connections(graph,0,1), CellArray)
  @test isa(connections(graph,0,2), CellArray)
  @test isa(connections(graph,0,3), CellArray)

end

@testset "CartesianDiscreteModel" begin

  model = CartesianDiscreteModel(
    domain=(0.0,1.0,-1.0,2.0,0.0,1.0),
    partition=(3,4,2))

  grid3 = Grid(model,3)

  @test isa(grid3,CartesianGrid{3})

  gridgraph = FullGridGraph(model)

  @test isa(gridgraph, FullGridGraph)

  grid2 = Grid(model,2)
  grid1 = Grid(model,1)
  grid0 = Grid(model,0)

  labels = FaceLabels(model)

  @test isa(labels,FaceLabels)
  @test length(labels_on_dim(labels,3)) == ncells(grid3)
  @test length(labels_on_dim(labels,2)) == ncells(grid2)
  @test length(labels_on_dim(labels,1)) == ncells(grid1)
  @test length(labels_on_dim(labels,0)) == ncells(grid0)

  @test ntags(labels) == 28
  @test name_from_tag(labels,28) == "boundary"

end

end # module GeometryTests
